#define BLOCKZ_SIZE 16
#define BLOCKX_SIZE 16 
#define BLOCKY_SIZE 16
#define FAT 4
#define COEFFS_SIZE 13

#define C0  0
#define CZ1 1
#define CX1 2
#define CY1 3
#define CZ2 4
#define CX2 5
#define CY2 6
#define CZ3 7
#define CX3 8
#define CY3 9
#define CZ4 10
#define CX4 11
#define CY4 12

#define C_C00(d) (8.0/(5.0*(d)*(d)))
#define get_coeffs(d1,d2,d3) {-1025.0/576.0*(C_C00(d1)+C_C00(d2)+C_C00(d3)), C_C00(d1), C_C00(d2), C_C00(d3),-C_C00(d1)/8.0,-C_C00(d2)/8.0,-C_C00(d3)/8.0,C_C00(d1)/63.0,C_C00(d2)/63.0,C_C00(d3)/63.0,-C_C00(d1)/896.0,-C_C00(d2)/896.0,-C_C00(d3)/896.0}

__constant__ float coeffs[COEFFS_SIZE];

/* FDTD kernel */
/*faster */
__global__ void wave_kernel(int nblocksz, float* vel, float *p0, float *p1, float *p2){
/* shared space */
	__shared__ float s_data[BLOCKX_SIZE+2*FAT][BLOCKZ_SIZE+2*FAT];

/* computing local(within thread block) and global(within global grid) coordinates */
	/* slow axis */
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;

/* fast axis(z) and intermediate axis (x) */
	int zl=FAT+threadIdx.x; 							// local  zdir
	int zg=(blockIdx.x-yblock*nblocksz)*blockDim.x+zl; 	// global zdir
	int xl=FAT+threadIdx.y; 							// local  xdir
	int xg=blockIdx.y*blockDim.y+xl; 					// global xdir

/* variables keeping slow axis(y) points within a thread */
	float 	p1y_p1, p1y_m1,
					p1y_p2, p1y_m2,
					p1y_p3, p1y_m3,    
					p1y_p4, p1y_m4;

/* boundary condition for fast and intermediate axes */
	int bc_temp=min2(min2(bc_zmax-zg,bc_xmax-xg),xg); 

/* temp. variables for faster addressing */
	int stride=n1gpu*n2gpu;
	/* mem. address for current point */
	int addr=zg+n1gpu*xg+stride*y0;
	int addr_forward=addr-stride-stride-stride-stride;

	/* init. y-axis (slow) variables */
	p1y_m3=p1[addr_forward];
	p1y_m2=p1[addr_forward+=stride];
	p1y_m1=p1[addr_forward+=stride];
	s_data[xl][zl]=p1[addr_forward+=stride];
	p1y_p1=p1[addr_forward+=stride];
	p1y_p2=p1[addr_forward+=stride];
	p1y_p3=p1[addr_forward+=stride];
	p1y_p4=p1[addr_forward+=stride];

	/* looping over the slow y-axis */
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){
		int y=y0+yl;

		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=s_data[xl][zl];	
		s_data[xl][zl]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1[addr_forward+=stride];

		/* boundary condition for slow axis */
		float bc=expf(bc_bgpu+min2(min2(y, bc_ymax-y),bc_temp)*bc_agpu);

/* loading into shared mem */
		if(threadIdx.x<FAT){ /* load xdir fat/halo */
			s_data[xl][threadIdx.x]=p1[addr-FAT];
			s_data[xl][zl+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
		}
		if(threadIdx.y<FAT){ /* load xdir fat/halo */
			s_data[threadIdx.y][zl]=p1[addr-FAT*n1gpu];
			s_data[xl+BLOCKX_SIZE][zl]=p1[addr+BLOCKX_SIZE*n1gpu];
		}
		__syncthreads();
		p2[addr]=
		vel[addr]*(
			 coeffs[C0]*s_data[xl][zl]
			+coeffs[CZ1]*(s_data[xl][zl+1]+s_data[xl][zl-1])
			+coeffs[CX1]*(s_data[xl+1][zl]+s_data[xl-1][zl])
			+coeffs[CY1]*(p1y_p1+p1y_m1)
			+coeffs[CZ2]*(s_data[xl][zl+2]+s_data[xl][zl-2])
			+coeffs[CX2]*(s_data[xl+2][zl]+s_data[xl-2][zl])
			+coeffs[CY2]*(p1y_p2+p1y_m2)
			+coeffs[CZ3]*(s_data[xl][zl+3]+s_data[xl][zl-3])
			+coeffs[CX3]*(s_data[xl+3][zl]+s_data[xl-3][zl])
			+coeffs[CY3]*(p1y_p3+p1y_m3)
			+coeffs[CZ4]*(s_data[xl][zl+4]+s_data[xl][zl-4])
			+coeffs[CX4]*(s_data[xl+4][zl]+s_data[xl-4][zl])
			+coeffs[CY4]*(p1y_p4+p1y_m4)
			+bc*(p0[addr]-s_data[xl][zl])
		)+s_data[xl][zl]+s_data[xl][zl]-p0[addr];
		addr+=stride;
	}
}

__global__ void wave_kernel4( float* vel, float *p0, float *p1, float *p2){
/* shared space */
	__shared__ float s_data[BLOCKX_SIZE+2*FAT][BLOCKZ_SIZE+2*FAT];

/* computing local(within thread block) and global(within global grid) coordinates */
	/* fast axis(z) and intermediate axis (x) */
	int zl=FAT+threadIdx.x; 							// local  zdir
	int zg=blockIdx.x*blockDim.x+zl; 	// global zdir
	int xl=FAT+threadIdx.y; 							// local  xdir
	int xg=blockIdx.y*blockDim.y+xl; 					// global xdir

/* variables keeping slow axis(y) points within a thread */
	register float 	p1y_p1, p1y_m1,
					p1y_p2, p1y_m2,
					p1y_p3, p1y_m3,    
					p1y_p4, p1y_m4;

/* boundary condition for fast and intermediate axes */
	int bc_temp=min2(min2(n1gpu-zg,n2gpu-xg),xg); 

/* temp. variables for faster addressing */
	register int stride=n1gpu*n2gpu;
	/* mem. address for current point */
	register int addr=zg+n1gpu*xg+stride*FAT;
	register int addr_forward=addr-stride-stride-stride-stride;

	/* init. y-axis (slow) variables */
	p1y_m3=p1[addr_forward];
	p1y_m2=p1[addr_forward+=stride];
	p1y_m1=p1[addr_forward+=stride];
	s_data[xl][zl]=p1[addr_forward+=stride];
	p1y_p1=p1[addr_forward+=stride];
	p1y_p2=p1[addr_forward+=stride];
	p1y_p3=p1[addr_forward+=stride];
	p1y_p4=p1[addr_forward+=stride];

	/* looping over the slow y-axis */
	for(int y=FAT; y<n3gpu-FAT; y++){

		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=s_data[xl][zl];	
		s_data[xl][zl]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1[addr_forward+=stride];

		/* boundary condition for slow axis */
		float bc=expf(bc_bgpu+min2(min2(y, n3gpu-y),bc_temp)*bc_agpu);

/* loading into shared mem */
		if(threadIdx.x<FAT){ /* load xdir fat/halo */
			s_data[xl][threadIdx.x]=p1[addr-FAT];
			s_data[xl][zl+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
		}
		if(threadIdx.y<FAT){ /* load xdir fat/halo */
			s_data[threadIdx.y][zl]=p1[addr-FAT*n1gpu];
			s_data[xl+BLOCKX_SIZE][zl]=p1[addr+BLOCKX_SIZE*n1gpu];
		}
		__syncthreads();
		p2[addr]=
		vel[addr]*(
			 coeffs[C0]*s_data[xl][zl]
			+coeffs[CZ1]*(s_data[xl][zl+1]+s_data[xl][zl-1])
			+coeffs[CX1]*(s_data[xl+1][zl]+s_data[xl-1][zl])
			+coeffs[CY1]*(p1y_p1+p1y_m1)
			+coeffs[CZ2]*(s_data[xl][zl+2]+s_data[xl][zl-2])
			+coeffs[CX2]*(s_data[xl+2][zl]+s_data[xl-2][zl])
			+coeffs[CY2]*(p1y_p2+p1y_m2)
			+coeffs[CZ3]*(s_data[xl][zl+3]+s_data[xl][zl-3])
			+coeffs[CX3]*(s_data[xl+3][zl]+s_data[xl-3][zl])
			+coeffs[CY3]*(p1y_p3+p1y_m3)
			+coeffs[CZ4]*(s_data[xl][zl+4]+s_data[xl][zl-4])
			+coeffs[CX4]*(s_data[xl+4][zl]+s_data[xl-4][zl])
			+coeffs[CY4]*(p1y_p4+p1y_m4)
			+bc*(p0[addr]-s_data[xl][zl])
		)+s_data[xl][zl]+s_data[xl][zl]-p0[addr];
		addr+=stride;
	}
}


__global__ void wave_kernel2(int nblocksz, float* vel, float *p0, float *p1, float *p2){

/* computing local(within thread block) and global(within global grid) coordinates */
	/* slow axis */
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;

/* fast axis(z) and intermediate axis (x) */
	int zl=FAT+threadIdx.x; 							// local  zdir
	int zg=(blockIdx.x-yblock*nblocksz)*blockDim.x+zl; 	// global zdir
	int xl=FAT+threadIdx.y; 							// local  xdir
	int xg=blockIdx.y*blockDim.y+xl; 					// global xdir

/* variables keeping slow axis(y) points within a thread */
	float 	p1y_p4, p1y_p3, p1y_p2, p1y_p1, p1y_center,
					p1y_m4, p1y_m3, p1y_m2, p1y_m1;

/* boundary condition for fast and intermediate axes */
	int bc_temp=min2(min2(n1gpu-zg,n2gpu-xg),xg); 

/* temp. variables for faster addressing */
	int stride=n1gpu*n2gpu;
	/* mem. address for current point */
	int addr=zg+n1gpu*xg+stride*y0;
	int addr_forward=addr-stride-stride-stride-stride;

	/* init. y-axis (slow) variables */
	p1y_m3=p1[addr_forward];
	p1y_m2=p1[addr_forward+=stride];
	p1y_m1=p1[addr_forward+=stride];
	p1y_center=p1[addr_forward+=stride];
	p1y_p1=p1[addr_forward+=stride];
	p1y_p2=p1[addr_forward+=stride];
	p1y_p3=p1[addr_forward+=stride];
	p1y_p4=p1[addr_forward+=stride];

	/* temps to prevent repetative address computation */
	int n1_x2=n1gpu+n1gpu;
	int n1_x3=n1_x2+n1gpu;
	int n1_x4=n1_x3+n1gpu;

	for(int yl=0; yl<BLOCKY_SIZE; ++yl){
		int y=y0+yl;
		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=p1y_center;	
		p1y_center=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1[addr_forward+=stride];

		/* boundary condition for slow axis */
		float bc=expf(bc_bgpu+min2(min2(y, n3gpu-y),bc_temp)*bc_agpu);

		p2[addr]=
		vel[addr]*(
			 coeffs[C0]*p1y_center
			+coeffs[CZ1]*(p1[addr+1]+p1[addr-1])
			+coeffs[CX1]*(p1[addr+n1gpu]+p1[addr-n1gpu])
			+coeffs[CY1]*(p1y_p1+p1y_m1)
			+coeffs[CZ2]*(p1[addr+2]+p1[addr-2])
			+coeffs[CX2]*(p1[addr+n1_x2]+p1[addr-n1_x2])
			+coeffs[CY2]*(p1y_p2+p1y_m2)
			+coeffs[CZ3]*(p1[addr+3]+p1[addr-3])
			+coeffs[CX3]*(p1[addr+n1_x3]+p1[addr-n1_x3])
			+coeffs[CY3]*(p1y_p3+p1y_m3)
			+coeffs[CZ4]*(p1[addr+4]+p1[addr-4])
			+coeffs[CX4]*(p1[addr+n1_x4]+p1[addr-n1_x4])
			+coeffs[CY4]*(p1y_p4+p1y_m4)
			+bc*(p0[addr]-p1y_center)
		)+p1y_center+p1y_center-p0[addr];
		addr+=stride;
	}
}

__global__ void wave_kernel1(int nblocksz, float* vel, float *p0, float *p1, float *p2){

	int z=FAT+(blockIdx.x%nblocksz)*blockDim.x+threadIdx.x; // global zdir
	int x=FAT+blockIdx.y*blockDim.y+threadIdx.y; // global xdir
	int y0=FAT+(blockIdx.x/nblocksz)*BLOCKY_SIZE;
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){
		int y=y0+yl;
		int addr=dloc(z,x,y);
		float bc=expf(bc_bgpu+min5(n1gpu-z,x,n2gpu-x,y, n3gpu-y)*bc_agpu);
		p2[addr]=
		vel[addr]*(
			 coeffs[C0]*p1[addr]
			+coeffs[CZ1]*(p1[dloc(z+1,x,y)]+p1[dloc(z-1,x,y)])
			+coeffs[CX1]*(p1[dloc(z,x+1,y)]+p1[dloc(z,x-1,y)])
			+coeffs[CY1]*(p1[dloc(z,x,y+1)]+p1[dloc(z,x,y-1)])
			+coeffs[CZ2]*(p1[dloc(z+2,x,y)]+p1[dloc(z-2,x,y)])
			+coeffs[CX2]*(p1[dloc(z,x+2,y)]+p1[dloc(z,x-2,y)])
			+coeffs[CY2]*(p1[dloc(z,x,y+2)]+p1[dloc(z,x,y-2)])
			+coeffs[CZ3]*(p1[dloc(z+3,x,y)]+p1[dloc(z-3,x,y)])
			+coeffs[CX3]*(p1[dloc(z,x+3,y)]+p1[dloc(z,x-3,y)])
			+coeffs[CY3]*(p1[dloc(z,x,y+3)]+p1[dloc(z,x,y-3)])
			+coeffs[CZ4]*(p1[dloc(z+4,x,y)]+p1[dloc(z-4,x,y)])
			+coeffs[CX4]*(p1[dloc(z,x+4,y)]+p1[dloc(z,x-4,y)])
			+coeffs[CY4]*(p1[dloc(z,x,y+4)]+p1[dloc(z,x,y-4)])
			+bc*(p0[addr]-p1[addr])
		)+2*p1[addr]-p0[addr];
	}
}
/* old */
__global__ void wave_kernel0( float* vel, float *p0, float *p1, float *p2){

	int z=FAT+blockIdx.x*blockDim.x+threadIdx.x; // global zdir
	int x=FAT+blockIdx.y*blockDim.y+threadIdx.y; // global xdir
	int y0=FAT;
	int ymax=(n3gpu-2*FAT);
	for(int yl=0; yl<ymax; ++yl){
		int y=y0+yl;
		int addr=dloc(z,x,y);
		float bc=expf(bc_bgpu+min5(n1gpu-z,x,n2gpu-x,y, n3gpu-y)*bc_agpu);
		p2[addr]=
		vel[addr]*(
			 coeffs[C0]*p1[addr]
			+coeffs[CZ1]*(p1[dloc(z+1,x,y)]+p1[dloc(z-1,x,y)])
			+coeffs[CX1]*(p1[dloc(z,x+1,y)]+p1[dloc(z,x-1,y)])
			+coeffs[CY1]*(p1[dloc(z,x,y+1)]+p1[dloc(z,x,y-1)])
			+coeffs[CZ2]*(p1[dloc(z+2,x,y)]+p1[dloc(z-2,x,y)])
			+coeffs[CX2]*(p1[dloc(z,x+2,y)]+p1[dloc(z,x-2,y)])
			+coeffs[CY2]*(p1[dloc(z,x,y+2)]+p1[dloc(z,x,y-2)])
			+coeffs[CZ3]*(p1[dloc(z+3,x,y)]+p1[dloc(z-3,x,y)])
			+coeffs[CX3]*(p1[dloc(z,x+3,y)]+p1[dloc(z,x-3,y)])
			+coeffs[CY3]*(p1[dloc(z,x,y+3)]+p1[dloc(z,x,y-3)])
			+coeffs[CZ4]*(p1[dloc(z+4,x,y)]+p1[dloc(z-4,x,y)])
			+coeffs[CX4]*(p1[dloc(z,x+4,y)]+p1[dloc(z,x-4,y)])
			+coeffs[CY4]*(p1[dloc(z,x,y+4)]+p1[dloc(z,x,y-4)])
			+bc*(p0[addr]-p1[addr])
		)+2*p1[addr]-p0[addr];
	}
}
